#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:     BayesianTracker
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
#-------------------------------------------------------------------------------

__author__ = "Alan R. Lowe"
__email__ = "a.lowe@ucl.ac.uk"

import logging
from . import hypothesis

from btrack.constants import Fates

from cvxopt.glpk import ilp
from cvxopt import matrix, spmatrix

# get the logger instance
logger = logging.getLogger('worker_process')

INIT_FATES = (Fates.INITIALIZE,
              Fates.INITIALIZE_BORDER,
              Fates.INITIALIZE_FRONT,
              Fates.INITIALIZE_LAZY)
TERM_FATES = (Fates.TERMINATE,
              Fates.TERMINATE_BORDER,
              Fates.TERMINATE_BACK,
              Fates.TERMINATE_LAZY)

class TrackOptimiser:
    """ TrackOptimiser

    TrackOptimiser takes a list of tracklets, as producing by a tracking
    algorithm, such as BayesianTracker, and tries to resolve small linkage
    errors, finding a globally optimal solution.

    General types of linkage error to resolve:
        Track lost at periphery -> Terminate the track
        Track lost briefly, but same state -> Link tracklets together by merging
        Object splitting (eg cell division) -> Do not merge, but create graph

    The algorithm proceeds as follows:
        1. Calculate the hypotheses for each tracklet
        2. Solve the association problem/globally optimise
        3. Merge/'link' trajectories
        4. Assign a 'fate' or optimal hypothesis to each trajectory

    The global optimisation utilises integer optimisation from the GLPK library.
    Takes in the list of hypotheses and formulates as a mixed integer linear
    programming problem.  Then attempts to use GLPK to solve the association
    problem, returns a list of hypotheses to act upon.

    We set up a constraints matrix, A in this manner: (num_hypotheses x 2N)
    Uses cvxopt.gplk.ilp to:

    minimize    c'*x
    subject to  G*x <= h
                A*x = b
                x[I] are all integer (empty set in this case)
                x[B] are all binary

    Since this is a *minimisation* we need to invert the (log) probability of
    each hypothesis.

    Args:
        hypotheses: a list of PyHypothesis objects from the tracker

    Members:
        optimise()

    Returns:
        a list of selected hypotheses. These are indices into the list of
        PyHypothesis objects.

    Notes:
        If the hypotheses are badly formed, this can take *FOREVER*.

        'Report Automated Cell Lineage Construction' Al-Kofahi et al.
        Cell Cycle 2006 vol. 5 (3) pp. 327-335

        'Reliable cell tracking by global data association', Bise et al.
        2011 IEEE Symposium on Biomedical Imaging pp. 1004-1010

        'Local cellular neighbourhood controls proliferation in cell
        competition', Bove A, Gradeci D, Fujita Y, Banerjee S, Charras G and
        Lowe AR 2017 Mol. Biol. Cell vol 28 pp. 3215-3228
    """

    def __init__(self):
        self._hypotheses = []

    @property
    def hypotheses(self):
        return self._hypotheses
    @hypotheses.setter
    def hypotheses(self, hypotheses):
        self._hypotheses = hypotheses

    def optimise(self):
        """
        We set up a constraints matrix, A in this manner: (num_hypotheses x 2N).
        Rho is the log probability of accepting the hypothesis. x is the
        integer set of hypotheses selected.

        Since this is a *minimisation* we need to invert the (log) probability
        of each hypothesis.
        """

        logger.info('Setting up constraints matrix for global optimisation...')

        # anon function to renumber track ID from C++
        trk_idx = lambda h: int(h)-1

        # calculate the number of hypotheses, could use this moment to cull?
        n_hypotheses = len(self.hypotheses)
        N = max(set([int(h.ID) for h in self.hypotheses]))

        # A is the constraints matrix (store as sparse since mostly empty)
        # note that we make this in the already transposed form...
        A = spmatrix([], [], [], (2*N, n_hypotheses), 'd')
        rho = matrix(0., (n_hypotheses, 1), 'd')

        # iterate over the hypotheses and build the constraints
        # TODO(arl): vectorize this for increased performance
        for counter, h in enumerate(self.hypotheses):

            # set the hypothesis score
            rho[counter] = h.log_likelihood

            if h.type == Fates.FALSE_POSITIVE:
                # is this a false positive?
                trk = trk_idx(h.ID)
                A[trk,counter] = 1
                A[N+trk,counter] = 1
                continue

            elif h.type in INIT_FATES:
                # an initialisation, therefore we only present this in the
                # second half of the A matrix
                trk = trk_idx(h.ID)
                A[N+trk,counter] = 1
                continue

            elif h.type in TERM_FATES:
                # a termination event, entry in first half only
                trk = trk_idx(h.ID)
                A[trk,counter] = 1
                continue

            elif h.type == Fates.APOPTOSIS:
                # an apoptosis event, entry in first half only
                trk = trk_idx(h.ID)
                A[trk,counter] = 1
                # A[N+trk,counter] = 1    # NOTE(arl): added 2019/08/29
                continue

            elif h.type == Fates.LINK:
                # a linkage event
                trk_i = trk_idx(h.ID)
                trk_j = trk_idx(h.link_ID)
                A[trk_i,counter] = 1
                A[N+trk_j,counter] = 1
                continue

            elif h.type == Fates.DIVIDE:
                # a branch event
                trk = trk_idx(h.ID)
                child_one = trk_idx(h.child_one_ID)
                child_two = trk_idx(h.child_two_ID)
                A[trk,counter] = 1
                A[N+child_one,counter] = 1
                A[N+child_two,counter] = 1
                continue

            elif h.type == Fates.MERGE:
                # a merge event
                trk = trk_idx(h.ID)
                parent_one = trk_idx(h.parent_one_ID)
                parent_two = trk_idx(h.parent_two_ID)
                A[N+trk,counter] = 1
                A[parent_one,counter] = 1
                A[parent_two,counter] = 1
                continue

            else:
                raise ValueError(f'Unknown hypothesis: {h.type}')

        logger.info('Optimizing...')

        # now set up the ILP solver
        G = spmatrix([], [], [], (2*N, n_hypotheses), 'd')
        h = matrix(0., (2*N,1), 'd')      # NOTE h cannot be a sparse matrix
        I = set()                         # empty set of x which are integer
        B = set(range(n_hypotheses))      # signifies all are binary in x
        b = matrix(1., (2*N,1), 'd')

        # now try to solve it!!!
        status, x = ilp(-rho, -G, h, A, b, I, B)

        # log the warning if not optimal solution
        if status != 'optimal':
            logger.warning(f'Optimizer returned status: {status}')
            return []

        # return only the selected hypotheses
        results = [i for i,h in enumerate(self.hypotheses) if x[i]>0]

        logger.info(f'Optimization complete. (Solution: {status})')
        return results


if __name__ == '__main__':
    pass



#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:     BayesianTracker
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
#-------------------------------------------------------------------------------

__author__ = "Alan R. Lowe"
__email__ = "a.lowe@ucl.ac.uk"

import os
import numpy as np
import ctypes
import json

from btrack import constants

H_TYPES = ['P_FP','P_init','P_term','P_link','P_branch','P_dead','P_merge']


class Hypothesis(ctypes.Structure):
    """ Hypothesis structure

      This is essentially a replicate of the C-structure used by the hypothesis
      engine. Added a few extra functions to enable quick assignment.

      unsigned int hypothesis;
      unsigned int ID;
      double probability;
      unsigned int link_ID;
      unsigned int child_one_ID;
      unsigned int child_two_ID;

      Notes:
        These are automatically generated by the optimiser
    """
    _fields_ = [('hypothesis',ctypes.c_uint),
                ('ID',ctypes.c_uint),
                ('probability',ctypes.c_double),
                ('link_ID', ctypes.c_uint),
                ('child_one_ID', ctypes.c_uint),
                ('child_two_ID', ctypes.c_uint),
                ('parent_one_ID', ctypes.c_uint),
                ('parent_two_ID', ctypes.c_uint)]

    @property
    def type(self):
        # return H_TYPES[self.hypothesis]
        return constants.Fates(self.hypothesis)

    @property
    def score(self):
        raise DeprecationWarning('score is deprecated. Use log_likelihood now.')

    @property
    def log_likelihood(self):
        return self.probability




class PyHypothesisParams(ctypes.Structure):
    """ HypothesisParams

    A convenient way to send hypothesis generation parameters to the hypothesis
    engine.

      double lambda_time;   // lambda_1
      double lambda_dist;   // lambda_2
      double lambda_link;   // lambda_3
      double lambda_branch; // lambda_4
      double eta;           // default low probability
      double theta_dist;
      double theta_time;
      double dist_thresh;
      double time_thresh;
      unsigned int apop_thresh;
      double segmentation_miss_rate;
      double apoptosis_rate;
      bool relax;

    Notes:
        None

    """
    _fields_ = [('lambda_time', ctypes.c_double),
                ('lambda_dist', ctypes.c_double),
                ('lambda_link', ctypes.c_double),
                ('lambda_branch', ctypes.c_double),
                ('eta', ctypes.c_double),
                ('theta_dist', ctypes.c_double),
                ('theta_time', ctypes.c_double),
                ('dist_thresh', ctypes.c_double),
                ('time_thresh', ctypes.c_double),
                ('apop_thresh', ctypes.c_uint),
                ('segmentation_miss_rate', ctypes.c_double),
                ('apoptosis_rate', ctypes.c_double),
                ('relax', ctypes.c_bool),
                ('hypotheses_to_generate', ctypes.c_uint)]

    def __init__(self, name=None):
        self.name = name

    @staticmethod
    def load(filename):
        return read_hypothesis_model(filename)






# def read_hypothesis_model(filename):
def read_hypothesis_model(config):
    """ read_hypothesis_model

    Read in a set of hypothesis parameters from a JSON description file.  The
    JSON file should contain the parameters of the PyHypothesisParams structure,
    and the function will return an instantiated PyHypothesisParams to be
    passed to the optimisation engine.

    Args:
        filename: the filename of the parameter file

    Notes:
        None
    """

    h_params = PyHypothesisParams()
    fields = [f[0] for f in h_params._fields_]

    # with open(filename, 'r') as json_file:
    #     config = json.load(json_file)

    for p in config['HypothesisModel']:
        if p in fields:
            setattr(h_params, p, config['HypothesisModel'][p])

    h_params.name = config['HypothesisModel']['name']

    # finally, take the hypotheses and setup a mask to generate only the
    # specified hypotheses
    # TODO(arl): no need to specify FP, this should be always be generated
    hypotheses = config['HypothesisModel']['hypotheses']
    h_bin = ''.join([str(int(h)) for h in [h in hypotheses for h in H_TYPES]])
    h_params.hypotheses_to_generate = int(h_bin[::-1], 2)

    return h_params







if __name__ == '__main__':
    pass



#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:     BayesianTracker
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
#-------------------------------------------------------------------------------


__author__ = "Alan R. Lowe"
__email__ = "a.lowe@ucl.ac.uk"

import os
import platform
import ctypes
import logging
import numpy as np

from . import utils

from .constants import BTRACK_PATH
from .btypes import PyTrackObject
from .btypes import PyTrackingInfo
from .optimise import hypothesis



# get the logger instance
logger = logging.getLogger('worker_process')


def numpy_pointer_decorator(func):
    """ simple decorator for numpy ctypes pointers """
    return func()

@numpy_pointer_decorator
def np_dbl_p():
    """ Temporary function. Will remove in final release """
    return np.ctypeslib.ndpointer(dtype=np.double, ndim=2, flags='C_CONTIGUOUS')

@numpy_pointer_decorator
def np_dbl_pc():
    """ Temporary function. Will remove in final release """
    return np.ctypeslib.ndpointer(dtype=np.double, ndim=2, flags='F_CONTIGUOUS')

@numpy_pointer_decorator
def np_uint_p():
    """ Temporary function. Will remove in final release """
    return np.ctypeslib.ndpointer(dtype=np.uint32, ndim=2, flags='C_CONTIGUOUS')

@numpy_pointer_decorator
def np_int_p():
    """ Temporary function. Will remove in final release """
    return np.ctypeslib.ndpointer(dtype=np.int32, ndim=2, flags='C_CONTIGUOUS')

@numpy_pointer_decorator
def np_int_vec_p():
    """ Temporary function. Will remove in final release """
    return np.ctypeslib.ndpointer(dtype=np.int32, ndim=1) #, flags='C_CONTIGUOUS')



def load_library(filename):
    """ Return the platform for shared library loading.  Take care of loading
    the appropriate shared library.

    Args:
        filename: filename for the library

    Raises:
        logging warning if windows is used
    """

    if not isinstance(filename, str):
        raise TypeError('Filename must be a string')

    lib_file, ext = os.path.splitext(filename)

    system = platform.system()
    version = platform.version()
    release = platform.release()

    file_ext = {'Linux':'.so', 'Darwin':'.dylib', 'Windows':'.DLL'}
    full_lib_file = lib_file + file_ext[system]

    try:
        lib = ctypes.cdll.LoadLibrary(full_lib_file)
        logger.info(f'Loaded btrack: {full_lib_file}')
    except IOError:
        raise IOError(f'Cannot load shared library {full_lib_file}')

    return lib



def get_library():
    """ This loads and returns the btrack shared library. """
    lib = load_library(os.path.join(BTRACK_PATH, 'libs', 'libtracker'))

    # deal with constructors/destructors
    lib.new_interface.restype = ctypes.c_void_p
    lib.new_interface.argtypes = [ctypes.c_bool]

    lib.del_interface.restype = None
    lib.del_interface.argtypes = [ctypes.c_void_p]

    # check the version number
    lib.check_library_version.restype = ctypes.c_bool
    lib.check_library_version.argtypes = [ctypes.c_uint, ctypes.c_uint,
                                          ctypes.c_uint]

    # set the update method
    lib.set_update_mode.restype = None
    lib.set_update_mode.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # set the motion model
    lib.motion.restype = None
    lib.motion.argtypes = [ctypes.c_void_p, ctypes.c_uint, ctypes.c_uint,
                            np_dbl_p, np_dbl_p, np_dbl_p, np_dbl_p,
                            np_dbl_p, ctypes.c_double, ctypes.c_double,
                            ctypes.c_uint, ctypes.c_double]

    # set the object model
    # lib.model.restype = None
    # lib.model.argtypes = [ctypes.c_void_p, ctypes.c_uint, np_dbl_p,
    #                         np_dbl_p, np_dbl_p]

    lib.max_search_radius.restype = None
    lib.max_search_radius.argtypes = [ctypes.c_void_p, ctypes.c_float]

    # append a new observation
    lib.append.restype = None
    lib.append.argtypes = [ctypes.c_void_p, PyTrackObject]

    # run the complete tracking
    lib.track.restype = ctypes.POINTER(PyTrackingInfo)
    lib.track.argtypes = [ctypes.c_void_p]

    # run one or more steps of the tracking, interactive mode
    lib.step.restype = ctypes.POINTER(PyTrackingInfo)
    lib.step.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get an individual track length
    lib.track_length.restype = ctypes.c_uint
    lib.track_length.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get a track
    lib.get.restype = ctypes.c_uint
    lib.get.argtypes = [ctypes.c_void_p, np_dbl_p, ctypes.c_uint]

    # get the internal ID of a track
    lib.get_ID.restype = ctypes.c_uint
    lib.get_ID.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get a track, by reference
    lib.get_refs.restype = ctypes.c_uint
    lib.get_refs.argtypes = [ctypes.c_void_p, np_int_vec_p, ctypes.c_uint]

    # get the parent ID (i.e. pre-division)
    lib.get_parent.restype = ctypes.c_uint
    lib.get_parent.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get the root ID (i.e. the root ID of this lineage tree)
    lib.get_root.restype = ctypes.c_uint
    lib.get_root.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get the ID of any children
    lib.get_children.restype = ctypes.c_uint
    lib.get_children.argtypes = [ctypes.c_void_p, np_int_vec_p, ctypes.c_uint]

    # get the fate of the track
    lib.get_fate.restype = ctypes.c_uint
    lib.get_fate.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get the generational depth of the track
    lib.get_generation.restype = ctypes.c_uint
    lib.get_generation.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get the kalman filtered position
    lib.get_kalman_mu.restype = ctypes.c_uint
    lib.get_kalman_mu.argtypes = [ctypes.c_void_p, np_dbl_p, ctypes.c_uint]

    # get the kalman covariance
    lib.get_kalman_covar.restype = ctypes.c_uint
    lib.get_kalman_covar.argtypes = [ctypes.c_void_p, np_dbl_p, ctypes.c_uint]

    # get the predicted position at each time step
    lib.get_kalman_pred.restype = ctypes.c_uint
    lib.get_kalman_pred.argtypes = [ctypes.c_void_p, np_dbl_p, ctypes.c_uint]

    # get the label of the object
    lib.get_label.restype = ctypes.c_uint
    lib.get_label.argtypes = [ctypes.c_void_p, np_uint_p, ctypes.c_uint]

    # get the imaging volume
    lib.get_volume.restype = None
    lib.get_volume.argtypes = [ctypes.c_void_p, np_dbl_p]

    # get the imaging volume
    lib.set_volume.restype = None
    lib.set_volume.argtypes = [ctypes.c_void_p, np_dbl_p]

    # return a dummy object by reference
    lib.get_dummy.restype = PyTrackObject
    lib.get_dummy.argtypes = [ctypes.c_void_p, ctypes.c_int]

    # get the number of tracks
    lib.size.restype = ctypes.c_uint
    lib.size.argtypes = [ctypes.c_void_p]

    # calculate the hypotheses
    lib.create_hypotheses.restype = ctypes.c_uint
    lib.create_hypotheses.argtypes = [ctypes.c_void_p,
                                      hypothesis.PyHypothesisParams,
                                      ctypes.c_uint, ctypes.c_uint]

    # get a hypothesis by ID
    lib.get_hypothesis.restype = hypothesis.Hypothesis
    lib.get_hypothesis.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # merge following optimisation
    lib.merge.restype = None
    lib.merge.argtypes = [ctypes.c_void_p, np_uint_p, ctypes.c_uint]

    return lib



import os
import enum

BTRACK_PATH = os.path.dirname(os.path.abspath(__file__))


def get_version():
    with open(os.path.join(BTRACK_PATH, "VERSION.txt") ,"r") as ver:
        version = ver.readline()
    return version.rstrip()

def get_version_tuple():
    return tuple([int(v) for v in get_version().split(".")])

DEFAULT_LOW_PROBABILITY = -1e5
MAX_LOST = 5
PROB_NOT_ASSIGN = 0.1
DEBUG = True
EXPORT_FORMATS = frozenset(['.json','.mat','.hdf5'])
NEW_COLORS = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b',
                '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
VOLUME = ((0,1024), (0,1024), (-100,100))
HDF_CHUNK_CACHE = 100*1024*1024
USER_MODEL_DIR = ""

DEFAULT_EXPORT_PROPERTIES = ['ID','t','x','y','z','parent','root', 'state',
                             'generation']


@enum.unique
class Errors(enum.Enum):
    SUCCESS = 900
    EMPTY_QUEUE = 901
    NO_TRACKS = 902
    NO_USEABLE_FRAMES = 903
    TRACK_EMPTY = 904
    INCORRECT_MOTION_MODEL = 905
    MAX_LOST_OUT_OF_RANGE = 906
    ACCURACY_OUT_OF_RANGE = 907
    PROB_NOT_ASSIGN_OUT_OF_RANGE = 908
    NOT_DEFINED = 909
    NO_ERROR = 910

@enum.unique
class Fates(enum.Enum):
    FALSE_POSITIVE = 0
    INITIALIZE = 1
    TERMINATE = 2
    LINK = 3
    DIVIDE = 4
    APOPTOSIS = 5
    MERGE = 6
    EXTRUDE = 7
    INITIALIZE_BORDER = 10
    INITIALIZE_FRONT = 11
    INITIALIZE_LAZY = 12
    TERMINATE_BORDER = 20
    TERMINATE_BACK = 21
    TERMINATE_LAZY = 22
    DEAD = 666
    UNDEFINED = 999

@enum.unique
class States(enum.Enum):
    INTERPHASE = 0
    PROMETAPHASE = 1
    METAPHASE = 2
    ANAPHASE = 3
    APOPTOSIS = 4
    NULL = 5
    DUMMY = 99

@enum.unique
class BayesianUpdates(enum.Enum):
    EXACT = 0
    APPROXIMATE = 1
    CUDA = 2



#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:     BayesianTracker
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
#-------------------------------------------------------------------------------


__author__ = "Alan R. Lowe"
__email__ = "a.lowe@ucl.ac.uk"

import ctypes
import numpy as np

from . import utils
from . import constants

# from datetime import datetime
from collections import OrderedDict
import itertools




class PyTrackObject(ctypes.Structure):
    """ TrackObject

    Primitive class to store information about an object. Essentially a single
    object in a field of view, with some member variables to keep track of data
    associated with an object.

    Args:
        position: 2D/3D position
        dummy: is this a real object or a dummy object (e.g. when lost)
        label: object classification
        attributes: object attributes, essentially metadata about object

    Properties:
        probability: class label probabilities
    """

    _fields_ = [('ID', ctypes.c_uint),
                ('x', ctypes.c_double),
                ('y', ctypes.c_double),
                ('z', ctypes.c_double),
                ('t', ctypes.c_uint),
                ('dummy', ctypes.c_bool),
                ('states', ctypes.c_uint),
                ('label', ctypes.c_int),
                ('prob', ctypes.c_double)]
                # ('prob', ctypes.POINTER(ctypes.c_double))]

    def __init__(self):
        super(PyTrackObject, self).__init__()
        self._raw_probability = None

    @property
    def probability(self):
        return self._raw_probability

    @probability.setter
    def probability(self, probability):
        if not isinstance(probability, np.ndarray):
            raise TypeError('.probability should be a numpy array')
        self._raw_probability = probability

    @property
    def state(self):
        return constants.States(self.label)




class PyTrackingInfo(ctypes.Structure):
    """ PyTrackingInfo

    Primitive class to store information about the tracking output.

    Params:
        error: error code from the tracker
        n_tracks: total number of tracks initialised during tracking
        n_active: number of active tracks
        n_conflicts: number of conflicts
        n_lost: number of lost tracks
        t_update_belief: time to update belief matrix in ms
        t_update_link: time to update links in ms
        t_total_time: total time to track objects
        p_link: typical probability of association
        p_lost: typical probability of losing track

    Notes:
        TODO(arl): should update to give more useful statistics, perhaps
        histogram of probabilities and timings.

    """

    _fields_ = [('error', ctypes.c_uint),
                ('n_tracks', ctypes.c_uint),
                ('n_active', ctypes.c_uint),
                ('n_conflicts', ctypes.c_uint),
                ('n_lost', ctypes.c_uint),
                ('t_update_belief', ctypes.c_float),
                ('t_update_link', ctypes.c_float),
                ('t_total_time', ctypes.c_float),
                ('p_link', ctypes.c_float),
                ('p_lost', ctypes.c_float),
                ('complete', ctypes.c_bool)]


    def to_dict(self):
        """ Return a dictionary of the statistics """
        # TODO(arl): make this more readable by converting seconds, milliseconds
        # and interpreting error messages?
        stats = {k:getattr(self, k) for k,typ in PyTrackingInfo._fields_}
        return stats

    @property
    def tracker_active(self):
        """ return the current status """
        no_error = constants.Errors(self.error) == constants.Errors.NO_ERROR
        return no_error and not self.complete



class MotionModel:
    """ MotionModel

    Kalman filter:
    'Is an algorithm which uses a series of measurements observed over time,
    containing noise (random variations) and other inaccuracies, and produces
    estimates of unknown variables that tend to be more precise than those that
    would be based on a single measurement alone.'

    Args:
        name: a name identifier
        measurements: the number of measurements of the system (typically x,y,z)
        states: the number of states of the system (typically >=measurements)
        A: State transition matrix
        B: Control matrix
        H: Observation matrix
        P: Initial covariance estimate
        Q: Estimated error in process
        R: Estimated error in measurements
        accuracy: integration limits for calculating the probabilities
        dt: time difference (always 1?)
        max_lost: number of frames without observation before marking as lost
        prob_not_assign: the default probability to not assign a track

    Members:
        reshape(): reshape matrices to the correct dimensions, will throw error
            if they are incorrectly sized.
        load(): load a motion model from a JSON file.

    Notes:
        This is just a wrapper for the data with a few convenience functions
        thrown in. Matrices must be stored Fortran style, because Eigen uses
        column major and Numpy uses row major storage.

    References:
        'A new approach to linear filtering and prediction problems.'
        Kalman RE, 1960 Journal of Basic Engineering
    """

    def __init__(self):
        self.name = 'Default'
        self.A = None
        self.H = None
        self.P = None
        self.G = None
        self.R = None
        self.measurements = None
        self.states = None
        self.dt = 1
        self.accuracy = 2.
        self.max_lost = constants.MAX_LOST
        self.prob_not_assign = constants.PROB_NOT_ASSIGN

    @property
    def Q(self):
        """ Return a Q matrix from the G matrix. """
        return self.G.transpose() * self.G

    def reshape(self):
        """ Reshapes matrices to the correct dimensions. Only need to call this
        if loading a model from a JSON file.

        Notes:
            Internally:
                Eigen::Matrix<double, m, s> H;
                Eigen::Matrix<double, s, s> Q;
                Eigen::Matrix<double, s, s> P;
                Eigen::Matrix<double, m, m> R;

        """
        s = self.states
        m = self.measurements

        # if we have defined a model, restructure matrices to the correct shapes
        # do some parsing to check that the model is specified correctly
        if s and m:
            shapes = {'A':(s,s), 'H':(m,s), 'P':(s,s), 'R':(m,m)}
            for m_name in shapes:
                try:
                    m_array = getattr(self, m_name)
                    r_matrix = np.reshape(m_array, shapes[m_name], order='C')
                except ValueError:
                    raise ValueError('Matrx {0:s} is incorrecly specified. '
                        '({1:d} entries for {2:d}x{3:d} matrix.)'.format(m_name,
                        len(m_array), shapes[m_name][0],
                        shapes[m_name][1]))

                setattr(self, m_name, r_matrix)
        else:
            raise ValueError('Cannot reshape matrices as MotionModel is '
                            'uninitialised')

    @staticmethod
    def load(filename):
        """ Load a model from file """
        return utils.read_motion_model(filename)



class ObjectModel:
    """ ObjectModel

    This is a class to deal with state transitions in the object, essentially
    a Hidden Markov Model.  Makes an assumption that the states are all
    observable, but with noise.

    Args:
        emission: the emission probability matrix
        transition: transition probabilities
        start: initial probabilities

    """
    def __init__(self):
        self.emission = None
        self.transition = None
        self.start = None
        self.states = None

    def reshape(self):
        """ Reshapes matrices to the correct dimensions. Only need to call this
        if loading a model from a JSON file.

        Notes:
            Internally:
                Eigen::Matrix<double, s, s> emission;
                Eigen::Matrix<double, s, s> transition;
                Eigen::Matrix<double, s, 1> start;
        """
        if not self.states:
            raise ValueError('Cannot reshape matrices as ObjectModel is '
                            'uninitialised')
        s = self.states
        self.emission = np.reshape(self.emission, (s,s), order='C')
        self.transition = np.reshape(self.transition, (s,s), order='C')

    @staticmethod
    def load(filename):
        """ Load a model from file """
        return utils.read_object_model(filename)



class Tracklet:
    """ Tracklet

    Tracklet object for storing and updating linked lists of track objects.
    Forms the data structure for an individual tracklet.

    Track 'fates' are the selected hypotheses after optimization. Defined in
    constants.Fates

    Args:
        ID: unique identifier
        data: trajectory
        kalman: Kalman filter output
        labels: class labels for each object
        fate: the fate of the track

    Members:
        __len__: length of the trajectory in frames (including interpolated)
        lost: whether the track was lost during tracking
        merge: a merging function to stitch together two tracks
        labeller: a mapping function from an integer to another label type (str)

    Properties:
        x: x position
        y: y position
        z: z position
        parent: parent tracklet
        root: root tracklet if a branching tree (ie cell division)
        motion_model: typically a reference to a Kalman filter or motion model
        is_root: boolean flag to denote root track
        is_leaf: boolean flag to denote leaf track
        start: first time stamp of track
        stop: last time stamp of track

    """

    def __init__(self,
                 ID: int,
                 data: list,
                 parent=None,
                 children=[],
                 fate=constants.Fates.UNDEFINED):

        assert all([isinstance(o, PyTrackObject) for o in data])

        self.ID = ID
        self._data = data
        self._kalman = None

        self.root = None
        self.parent = parent
        self.children = children
        self.type = None
        self.fate = fate
        self.generation = 0

    def __len__(self):
        return len(self._data)

    @property
    def x(self): return [o.x for o in self._data]
    @property
    def y(self): return [o.y for o in self._data]
    @property
    def z(self): return [o.z for o in self._data]
    @property
    def t(self): return [o.t for o in self._data]
    @property
    def dummy(self): return [o.dummy for o in self._data]

    @property
    def start(self): return self.t[0]
    @property
    def stop(self): return self.t[-1]

    @property
    def label(self):
        return [o.state.name for o in self._data]

    @property
    def state(self):
        return [o.state.value for o in self._data]

    @property
    def softmax(self):
        return [o.probability for o in self._data]

    @property
    def is_root(self) -> bool:
        return self.parent == 0 or self.parent == None or self.parent == self.ID

    @property
    def is_leaf(self) -> bool:
        return not self.children

    @property
    def kalman(self): return self._kalman
    @kalman.setter
    def kalman(self, data):
        assert(isinstance(data, np.ndarray))
        self._kalman = data

    def mu(self, index):
        """ Return the Kalman filter mu. Note that we are only returning the mu
         for the positions (e.g. 3x1) """
        return np.matrix(self.kalman[index,1:4]).reshape(3,1)

    def covar(self, index):
        """ Return the Kalman filter covariance matrix. Note that we are
        only returning the covariance matrix for the positions (e.g. 3x3) """
        return np.matrix(self.kalman[index,4:13]).reshape(3,3)

    def predicted(self, index):
        """ Return the motion model prediction for the given timestep. """
        return np.matrix(self.kalman[index,13:]).reshape(3,1)

    def to_dict(self, properties: list = constants.DEFAULT_EXPORT_PROPERTIES):
        """ Return a dictionary of the tracklet which can be used for JSON
        export. This is an ordered dictionary for nicer JSON output.
        """
        trk_tuple = tuple([(p, getattr(self, p)) for p in properties])
        return OrderedDict(trk_tuple)

    def to_array(self, properties: list = constants.DEFAULT_EXPORT_PROPERTIES):
        """ Return a numpy array of the tracklet which can be used for MATLAB
        export. """

        tmp_track = np.zeros((len(self), len(properties)), dtype=np.float32)
        for i, property in enumerate(properties):
            tmp_track[:,i] = getattr(self, property)
        return tmp_track

    def in_frame(self, frame):
        """ Return true or false as to whether the track is in the frame """
        return self.t[0]<=frame and self.t[-1]>=frame

    def trim(self, frame, tail=75):
        """ Trim the tracklet and return one with the trimmed data """
        d = [o for o in self._data if o.t<=frame and o.t>=frame-tail]
        return Tracklet(self.ID, d)



import os
import numpy as np
import matplotlib.pyplot as plt
from .render import _render_tracks, _check_plot_order, _draw_cube

def _render_agents(ax, agents, colors_rgb=[], order='xyz', labels=False,
                   terminii=False, lw=5., frame=None):
    """ Render the agents """
    DIMS = _check_plot_order(order)
    if frame:
        p = [[getattr(a, order[i]) for a in agents if a.t==frame] for i in range(DIMS)]
    else:
        p = [[getattr(a, order[i]) for a in agents] for i in range(DIMS)]
    ax.scatter(*p, c='lightgrey', alpha=0.9)


class RenderTrackingMovie(object):
    """ RenderTrackingMovie

    Render tracks as a movie, by slicing tracks to show comet tails.

    Args:
        tracks: the tracks
        agents: (optional) agent objects if a simulation
        volume: the tracking volume

    Properties:
        save_dir: directory to save rendered images to

    Notes:
        TODO(arl): render image data
    """

    def __init__(self, tracks, agents, volume):
        self._tracks = tracks
        self._agents = agents
        self._volume = volume
        self.cmap = plt.get_cmap('gnuplot')

        self.image_stack = None

        self.xx, self.yy = np.meshgrid(np.linspace(0,1599,800), np.linspace(0,1199,600))

    @property
    def save_dir(self):
        return self._save_dir
    @save_dir.setter
    def save_dir(self, save_dir):
        if not isinstance(save_dir, str):
            raise TypeError('Save directory must be a string.')
        if not os.path.exists(save_dir):
            raise IOError('Save directory does not exist.')
        self._save_dir = save_dir


    def _display_image(self, ax, frame):
        im = self.image_stack[frame][0:-1:2,0:-1:2]
        X =  xx
        Y =  yy
        Z =  frame*np.ones(X.shape)
        ax.plot_surface(X, Y, Z, rstride=1, cstride=1, facecolors=im, shade=False)


    def render(self, order='xyz', labels=True, lw=3., tail=75):
        """ Render the movie """
        DIMS = 3 #_check_plot_order(order)

        fig = plt.figure(figsize=(16,16))

        # set up fiddly plot functions
        if DIMS == 3:
            ax = fig.add_subplot(111, projection='3d')
        else:
            ax = fig.add_subplot(111)

        colors_rgb = [self.cmap(int(i)) for i in np.linspace(0,255,16)]
        p_args = {'colors_rgb':colors_rgb, 'order':order, 'labels':labels}

        for f in range(self._volume[3][0], self._volume[3][1]):
            ax.cla()

            if self.image_stack is not None:
                self._display_image(ax, f)

            # _render_agents(ax, self._agents, lw=5., **p_args)
            to_plot = [ t.trim(f, tail=tail) for t in self._tracks if t.in_frame(f) ]
            _render_agents(ax, self._agents, frame=f)
            _render_tracks(ax, to_plot, lw=lw, **p_args)


            # plot a box of the appropriate dimensions
            if DIMS == 3:
                box_dims = 'xyzt'
                this_box = [self._volume[box_dims.index(dim)] for dim in order]
                _draw_cube(ax, this_box)

            ax.set_xlabel(order[0])
            ax.set_ylabel(order[1])
            if DIMS == 3:
                ax.set_zlabel(order[2])
            else:
                ax.autoscale('tight')

            ax.view_init(30, 150.+180.*(float(f)/float(self._volume[3][1])))

            plt.title('Tracking {0:d} objects in frame {1:d}'.format(len(to_plot), f))
            plt.axis('image')
            plt.savefig(os.path.join(self.save_dir,'tracking_'+str(f)+'.png'))

        plt.show()



#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:     BayesianTracker
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
#-------------------------------------------------------------------------------

__author__ = "Alan R. Lowe"
__email__ = "a.lowe@ucl.ac.uk"

import os
import json
import time
import ctypes
import logging
import numpy as np

from . import utils
from . import constants
from . import btypes
from . import libwrapper

from .dataio import export_delegator
from .optimise import optimiser

from datetime import datetime
from collections import OrderedDict
import itertools

__version__ = constants.get_version()

# get the logger instance
logger = logging.getLogger('worker_process')

# if we don't have any handlers, set one up
if not logger.handlers:
    # configure stream handler
    log_formatter = logging.Formatter('[%(levelname)s][%(asctime)s] %(message)s',
        datefmt='%Y/%m/%d %I:%M:%S %p')
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)

    logger.addHandler(console_handler)
    logger.setLevel(logging.DEBUG)





class BayesianTracker:
    """ BayesianTracker

    BayesianTracker is a multi object tracking algorithm, specifically
    used to reconstruct tracks in crowded fields. Here we use a probabilistic
    network of information to perform the trajectory linking. This method uses
    positional information (position, velocity ...) as well as visual
    information (labels, features...) for track linking.

    The tracking algorithm assembles reliable sections of track that do not
    contain splitting events (tracklets). Each new tracklet initiates a
    probabilistic model in the form of a Kalman filter (Kalman, 1960), and
    utilises this to predict future states (and error in states) of each of the
    objects in the field of view.  We assign new observations to the growing
    tracklets (linking) by evaluating the posterior probability of each
    potential linkage from a Bayesian belief matrix for all possible linkages
    (Narayana and Haverkamp, 2007). The best linkages are those with the highest
    posterior probability.

    This class is a wrapper for the C++ implementation of the BayesianTracker.

    Data can be passed in in the following formats:
        - btrack PyTrackObject (defined in btypes)
        - Optional JSON files using loaders
        - HDF

    Can be used with ContextManager support, like this:

        >>> with BayesianTracker() as tracker:
        >>>    tracker.append(observations)
        >>>    tracker.track()

    The tracker can be used to return all of the original data neatly packaged
    into tracklet objects, or as a nested list of references to the original
    data sets. The latter is useful if this is only the first part of a tracking
    protocol, or other metadata is needed for further analysis. The references
    can be used to make symbolic links in HDF5 files, for example.

    Use the .tracks to return Tracklets, or .refs to return the references.

    Use optimise to generate hypotheses for global optimisation. Read the
    TrackLinker documentation for more information about the track linker.

    Members:
        append(): append an object (or list of objects)
        track(): run the tracking algorithm
        track_interactive(): run the tracking in interactive mode
        optimise(): run the optimiser
        cleanup(): clean up the tracks according to some metrics
        configure_from_file(): pass a json configuration file

    Args:
        motion_model: a motion model to make motion predictions
        object_model: an object model to make state predictions
        return_kalman: boolean to request the Kalman debug info

    Properties:
        n_tracks: number of found tracks
        tracks: the tracks themselves
        refs: the tracks (by reference)
        dummies: the dummy objects inserted by the tracker
        volume: the imaging volume [x,y,z,t]
        frame_range: the frame range for tracking, essentially the last
            dimension of volume
        max_search_radius: maximum search radius when using fast cost update

    Notes:
        TODO(arl): lower precision for tracking output?

    References:
        'A Bayesian algorithm for tracking multiple moving objects in outdoor
        surveillance video', Narayana M and Haverkamp D 2007 IEEE

        'Report Automated Cell Lineage Construction' Al-Kofahi et al.
        Cell Cycle 2006 vol. 5 (3) pp. 327-335

        'Reliable cell tracking by global data association', Bise et al.
        2011 IEEE Symposium on Biomedical Imaging pp. 1004-1010

        'Local cellular neighbourhood controls proliferation in cell
        competition', Bove A, Gradeci D, Fujita Y, Banerjee S, Charras G and
        Lowe AR 2017 Mol. Biol. Cell vol 28 pp. 3215-3228
    """
    def __init__(self, verbose=True):
        """ Initialise the BayesianTracker C++ engine and parameters """

        # default parameters and space for stored objects
        self._motion_model = None
        self._object_model = None
        self._frame_range = [0,0]
        self._max_search_radius = np.inf
        self.return_kalman = False
        self._objects = []

        # do not initialise until the init() has been run
        self._initialised = False

        # load the library
        self._lib = libwrapper.get_library()

        # get an instance of the engine
        self._engine = self._lib.new_interface(verbose)

        # sanity check library version
        version_tuple = constants.get_version_tuple()
        if not self._lib.check_library_version(self._engine, *version_tuple):
            logger.warning(f'btrack (v{__version__}) shared library mismatch.')
        else:
            logger.info(f'btrack (v{__version__}) library imported')

        # silently set the update method to EXACT
        self._bayesian_update_method = constants.BayesianUpdates.EXACT
        self._lib.set_update_mode(self._engine, self.update_method.value)


    def __enter__(self):
        logger.info('Starting BayesianTracker session')
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        logger.info('Ending BayesianTracker session')
        self._lib.del_interface( self._engine )


    def configure_from_file(self, filename):
        """ Configure the tracker from a configuration file """
        config = utils.load_config(filename)
        self.configure(config)


    def configure(self, config):
        """ Configure the tracker with a motion model, an object model and
        hypothesis generation_parameters.
        """

        if not isinstance(config, dict):
            raise TypeError('configuration must be a dictionary')

        # store the models locally
        self.motion_model = config.get("MotionModel", None)
        self.object_model = config.get("ObjectModel", None)
        self.hypothesis_model = config.get("HypothesisModel", None)

        self._initialised = True


    def __len__(self): return self.n_tracks


    @property
    def max_search_radius(self):
        return self._max_search_radius
    @max_search_radius.setter
    def max_search_radius(self, max_search_radius):
        """ Set the maximum search radius for fast cost updates """
        assert(max_search_radius>0.)
        logger.info(f'Setting max XYZ search radius to: {max_search_radius}')
        self._lib.max_search_radius(self._engine, max_search_radius)


    @property
    def update_method(self):
        return self._bayesian_update_method
    @update_method.setter
    def update_method(self, method):
        """ set the method for updates, EXACT, APPROXIMATE, CUDA etc... """
        assert(method in constants.BayesianUpdates)
        logger.info(f'Setting Bayesian update method to: {method}')
        self._lib.set_update_mode(self._engine, method.value)
        self._bayesian_update_method = method


    @property
    def n_tracks(self):
        """ Return the number of tracks found """
        return self._lib.size( self._engine )


    @property
    def n_dummies(self):
        """ Return the number of dummy objects (negative ID) """
        return len([d for d in itertools.chain.from_iterable(self.refs) if d<0])


    @property
    def tracks(self):
        """ Return a sorted list of tracks, default is to sort by increasing
        length """
        return [self[i] for i in range(self.n_tracks)]


    @property
    def refs(self):
        """ Return tracks as a list of IDs (essentially pointers) to the
        original objects. Use this to write out HDF5 tracks. """
        tracks = []
        for i in range(self.n_tracks):
            # get the track length
            n = self._lib.track_length(self._engine, i)

            # set up some space for the output and  get the track data
            refs = np.zeros((n,),dtype='int32')
            _ = self._lib.get_refs(self._engine, refs, i)
            tracks.append(refs.tolist())

        return tracks


    @property
    def dummies(self):
        """ Return a list of dummy objects """
        return [self._lib.get_dummy(self._engine, -(i+1)) for i in range( self.n_dummies)]


    @property
    def lbep(self):
        """ Return an LBEP list
        > L - a unique label of the track (label of markers, 16-bit positive)
        > B - a zero-based temporal index of the frame in which the track begins
        > E - a zero-based temporal index of the frame in which the track ends
        > P - label of the parent track (0 is used when no parent is defined)
        > R - label of the root track
        > G - generational depth (from root)
        """
        lbep = lambda t: (t.ID, t.start, t.stop, t.parent, t.root, t.generation)
        return [lbep(t) for t in self.tracks]


    def _sort(self, tracks):
        """ Return a sorted list of tracks """
        return sorted(tracks, key=lambda t:len(t), reverse=True)


    @property
    def volume(self):
        """ Return the imaging volume in the format xyzt. This is effectively
        the range of each dimension: [(xlo,xhi), ..., (zlo,zhi), (tlo,thi)]
        """
        vol = np.zeros((3,2),dtype='float')
        self._lib.get_volume(self._engine, vol)
        return [tuple(vol[i,:].tolist()) for i in range(3)]+[self.frame_range]
    @volume.setter
    def volume(self, volume):
        """ Set the imaging volume """
        if not isinstance(volume, tuple):
            raise TypeError('Volume must be a tuple')
        if len(volume) != 3 or any([len(v)!=2 for v in volume]):
            raise ValueError('Volume must contain three tuples (xyz)')
        self._lib.set_volume(self._engine, np.array(volume, dtype='float64'))
        logger.info(f'Set volume to {volume}')


    @property
    def motion_model(self):
        return self._motion_model
    @motion_model.setter
    def motion_model(self, new_model):
        """ Set a new motion model. Must be of type MotionModel, either loaded
        from file or instantiating a MotionModel.

        Args:
            new_model: can be a string or a user defined MotionModel class

        Raises:
            TypeError is cannot determine the type of motion model
        """

        if isinstance(new_model, btypes.MotionModel):
            # TODO(arl): model parsing for a user defined model
            model = new_model
        else:
            raise TypeError('Motion model needs to be defined in /models/ or'
                'provided as a MotionModel object')

        self._motion_model = model
        logger.info(f'Loading motion model: {model.name}')

        # need to populate fields in the C++ library
        self._lib.motion( self._engine, model.measurements, model.states, model.A,
            model.H, model.P, model.Q, model.R, model.dt, model.accuracy,
            model.max_lost, model.prob_not_assign )


    @property
    def object_model(self):
        return self._object_model
    @object_model.setter
    def object_model(self, new_model):
        """
        Set a new object model. Must be of type ObjectModel, either loaded
        from file or instantiating an ObjectModel.

        Args:
            new_model: can be a string or a user defined ObjectModel class

        Raises:
            TypeError if cannot determine the type of motion model
        """

        if isinstance(new_model, btypes.ObjectModel):
            # this could be a user defined model
            # TODO(arl): model parsing
            model = new_model
        elif new_model is None:
            return
        else:
            raise TypeError('Object model needs to be defined in /models/ or'
            'provided as a ObjectModel object')

        self._object_model = model
        logger.info(f'Loading object model: {model.name}')

        # need to populate fields in the C++ library
        self._lib.model( self._engine, model.states, model.emission,
            model.transition, model.start )


    @property
    def frame_range(self):
        return self._frame_range
    @frame_range.setter
    def frame_range(self, frame_range):
        if not isinstance(frame_range, tuple):
            raise TypeError('Frame range must be specified as a tuple')
        if frame_range[1] < frame_range[0]:
            raise ValueError('Frame range must be low->high')
        self._frame_range = frame_range


    @property
    def objects(self):
        return self._objects

    def append(self, objects):
        """ Append a single track object, or list of objects to the stack. Note
        that the tracker will automatically order these by frame number, so the
        order here does not matter. This means several datasets can be
        concatenated easily, by running this a few times. """

        if not isinstance(objects, list):
            objects = [objects]

        for idx, obj in enumerate(objects):
            obj.ID = idx + len(self._objects) # make sure ID tracks properly
            if not isinstance(obj, btypes.PyTrackObject):
                raise TypeError('track_object must be a PyTrackObject')

            self._frame_range[1] = max(obj.t, self._frame_range[1])
            ret = self._lib.append( self._engine, obj )

        # store a copy of the list of objects
        self._objects += objects


    def _stats(self, info_ptr):
        """ Cast the info pointer back to an object """

        if not isinstance(info_ptr, ctypes.POINTER(btypes.PyTrackingInfo)):
            raise TypeError('Stats requires the pointer to the object')

        return info_ptr.contents


    def track(self):
        """ Run the actual tracking algorithm """

        if not self._initialised:
            logger.error('Tracker has not been configured')
            return

        logger.info('Starting tracking... ')
        # ret, tm = timeit( lib.track,  self._engine )
        ret = self._lib.track(self._engine)

        # get the statistics
        stats = self._stats(ret)

        if not utils.log_error(stats.error):
            logger.info(f('SUCCESS. Found {self.n_tracks} tracks in'
                         f'{1+self._frame_range[1]} frames'))

        # can log the statistics as well
        utils.log_stats(stats.to_dict())


    def track_interactive(self, step_size=100):
        """ Run the tracking in an interactive mode """

        # TODO(arl): this needs cleaning up to have some decent output
        if not self._initialised:
            logger.error('Tracker has not been configured')
            return

        logger.info('Starting tracking... ')

        stats = self.step()
        frm = 0

        # while not stats.complete and stats.error not in constants.ERRORS:
        while stats.tracker_active:
            logger.info((f'Tracking objects in frames {frm} to '
                         f'{min(frm+step_size-1, self._frame_range[1]+1)} '
                         f'(of {self._frame_range[1]+1})...'))

            stats = self.step(step_size)
            utils.log_stats(stats.to_dict())
            frm+=step_size

        if not utils.log_error(stats.error):
            logger.info('SUCCESS.')
            logger.info((f' - Found {self.n_tracks} tracks in '
                         f'{1+self._frame_range[1]} frames '
                         f'(in {stats.t_total_time}s)'))
            logger.info((f' - Inserted {self.n_dummies} dummy objects to fill '
                         'tracking gaps'))


    def step(self, n_steps=1):
        """ Run an iteration (or more) of the tracking. Mostly for
        interactive mode tracking """
        if not self._initialised: return None
        return self._stats(self._lib.step( self._engine, n_steps ))


    def hypotheses(self, params=None):
        """ Calculate and return hypotheses using the hypothesis engine """
        # raise NotImplementedError
        if not self.hypothesis_model:
            raise AttributeError('Hypothesis model has not been specified.')

        n_hypotheses = self._lib.create_hypotheses(self._engine,
            self.hypothesis_model, self.frame_range[0], self.frame_range[1])

        # now get all of the hypotheses
        h = [self._lib.get_hypothesis(self._engine, h) for h in range(n_hypotheses)]
        return h


    def optimize(self): return self.optimise()
    def optimise(self):
        """ Optimise the tracks. This generates the hypotheses for track merges,
        branching etc, runs the optimiser and then performs track merging,
        removal of track fragments, renumbering and assignment of branches.
        """

        logger.info(f'Loading hypothesis model: {self.hypothesis_model.name}')

        logger.info(f'Calculating hypotheses (relax: {self.hypothesis_model.relax})...')
        hypotheses = self.hypotheses()

        # if we don't have any hypotheses return
        if not hypotheses:
            logger.warning('No hypotheses could be found.')
            return []

        # set up the track optimiser
        track_linker = optimiser.TrackOptimiser()
        track_linker.hypotheses = hypotheses
        selected_hypotheses = track_linker.optimise()
        optimised = [hypotheses[i] for i in selected_hypotheses]

        h_original = [h.type for h in hypotheses]
        h_optimise = [h.type for h in optimised]
        h_types = sorted(list(set(h_original)), key=lambda h: h.value)

        for h_type in h_types:
            logger.info((f' - {h_type}: {h_optimise.count(h_type)}'
                         f' (of {h_original.count(h_type)})'))
        logger.info(f' - TOTAL: {len(hypotheses)} hypotheses')

        # now that we have generated the optimal sequence, merge all of the
        # tracks, delete fragments and assign divisions
        h_array = np.array(selected_hypotheses, dtype='uint32')
        h_array = h_array[np.newaxis,...]
        self._lib.merge(self._engine, h_array, len(selected_hypotheses))

        return optimised


    def __getitem__(self, idx):
        """ Grab a track from the BayesianTracker object. """
        # get the track length
        n = self._lib.track_length(self._engine, idx)

        # set up some space for the output
        children = np.zeros((2,), dtype=np.int32)    # pointers to children
        refs = np.zeros((n,), dtype=np.int32)        # pointers to objects

        # get the track data
        _ = self._lib.get_refs(self._engine, refs, idx)
        nc = self._lib.get_children(self._engine, children, idx)
        p = self._lib.get_parent(self._engine, idx)
        f = constants.Fates( self._lib.get_fate(self._engine, idx) )

        # get the track ID
        trk_id = self._lib.get_ID(self._engine, idx)

        # convert the array of children to a python list
        if nc > 0:
            c = children.tolist()
        else:
            c = []

        # now build the track from the references
        refs = refs.tolist()
        dummies = [self._lib.get_dummy(self._engine, d) for d in refs if d<0]

        track = []
        for r in refs:
            if r<0:
                # TODO(arl): softmax scores are zero for dummy objects
                dummy = dummies.pop(0)
                dummy.probability = np.zeros((5,), dtype=np.float32)
                track.append(dummy)
            else:
                track.append(self._objects[r])

        # make a new track object and return it
        trk = btypes.Tracklet(trk_id, track, parent=p, children=c, fate=f)
        trk.root = self._lib.get_root(self._engine, idx)
        trk.generation = self._lib.get_generation(self._engine, idx)

        if not self.return_kalman: return trk

        # get the size of the Kalman arrays
        sz_mu = self.motion_model.measurements + 1
        sz_cov = self.motion_model.measurements**2 + 1

        # otherwise grab the kalman filter data
        kal_mu = np.zeros((n, sz_mu), dtype=np.float32)     # kalman filtered
        kal_cov = np.zeros((n, sz_cov), dtype=np.float32)   # kalman covariance
        kal_pred = np.zeros((n, sz_mu), dtype=np.float32)   # motion predict

        n_kal = self._lib.get_kalman_mu(self._engine, kal_mu, idx)
        _ = self._lib.get_kalman_covar(self._engine, kal_cov, idx)
        _ = self._lib.get_kalman_pred(self._engine, kal_pred, idx)

        # cat the data [mu(0),...,mu(n),cov(0,0),...cov(n,n), pred(0),..]
        trk.kalman = np.hstack((kal_mu, kal_cov[:,1:], kal_pred[:,1:]))
        return trk


    def cleanup(self, sigma=2.5):
        """ Clean up following tracking. Can be used to remove static objects
        and links that are greater than the maximum distance permitted """
        dynamic_track = lambda trk: (np.std(trk.x)+np.std(trk.y))*0.5 > sigma
        return [t for t in self.tracks if len(t)>1 and dynamic_track(t)]

    def export(self, filename, obj_type=None, filter_by=None):
        """ export tracks using the appropriate exporter """
        export_delegator(filename, self, obj_type=obj_type, filter_by=filter_by)













if __name__ == "__main__":
    pass



#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:     BayesianTracker
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
#-------------------------------------------------------------------------------


__author__ = "Alan R. Lowe"
__email__ = "code@arlowe.co.uk"

import re
import os
import numpy as np
import time
import csv
import json

import logging

# import core
from . import btypes
from . import constants
from .optimise import hypothesis

from collections import OrderedDict
from scipy.io import savemat



# get the logger instance
logger = logging.getLogger('worker_process')


def load_config(filename):
    """ Load a tracking configuration file """
    if not os.path.exists(filename):
        # check whether it exists in the user model directory
        _, fn = os.path.split(filename)
        local_filename = os.path.join(constants.USER_MODEL_DIR, fn)

        if not os.path.exists(local_filename):
            logger.error(f'Configuration file {filename} not found')
            raise IOError(f'Configuration file {filename} not found')
        else:
            filename = local_filename

    with open(filename, 'r') as config_file:
        config = json.load(config_file)

    if "TrackerConfig" not in config:
        logger.error("Configuration file is malformed.")
        raise Exception("Tracking config is malformed")

    config = config["TrackerConfig"]

    logger.info(f'Loading configuration file: {filename}')
    t_config = {"MotionModel": read_motion_model(config),
                "ObjectModel": read_object_model(config),
                "HypothesisModel": hypothesis.read_hypothesis_model(config)}

    return t_config



def log_error(err_code):
    """ Take an error code from the tracker and log an error for the user. """
    error = constants.Errors(err_code)
    if error != constants.Errors.SUCCESS and error != constants.Errors.NO_ERROR:
        logger.error(f'ERROR: {error}')
        return True
    return False



def log_stats(stats):
    """ Take the statistics from the track and log the output """

    if log_error(stats['error']): return

    logger.info(' - Timing (Bayesian updates: {0:.2f}ms, Linking:'
                ' {1:.2f}ms)'.format(stats['t_update_belief'],
                stats['t_update_link']))

    logger.info(' - Probabilities (Link: {0:.5f}, Lost:'
                ' {1:.5f})'.format(stats['p_link'], stats['p_lost']))

    if stats['complete']:
        return

    logger.info(' - Stats (Active: {0:d}, Lost: {1:d}, Conflicts '
                'resolved: {2:d})'.format(stats['n_active'],
                stats['n_lost'], stats['n_conflicts']))



# def read_motion_model(filename):
def read_motion_model(config):
    """ read_motion_model

    Read in a motion model description file and return a dictionary containing
    the appropriate parameters.

    Motion models can be described using JSON format, with a basic structure
    as follows:

        {
          "MotionModel":{
            "name": "ConstantVelocity",
            "dt": 1.0,
            "measurements": 3,
            "states": 6,
            "accuracy": 2.0,
            "A": {
              "matrix": [1,0,0,1,0,0,...
              ...
              ] }
            }
        }

    Matrices are flattened JSON arrays.

    Most are self explanatory, except accuracy (perhaps a misnoma) - this
    represents the integration limits when determining the probabilities from
    the multivariate normal distribution.

    Args:
        filename: a JSON file describing the motion model.

    Returns:
        model: a btypes.MotionModel instance for passing to BayesianTracker

    Notes:
        Note that the matrices are stored as 1D matrices here. In the future,
        this could form part of a Python only motion model.

        TODO(arl): More parsing of the data/reshaping arrays. Raise an
        appropriate error if there is something wrong with the model definition.
    """
    matrices = frozenset(['A','H','P','G','R'])
    model = btypes.MotionModel()

    if 'MotionModel' not in list(config.keys()):
        raise ValueError('Not a valid motion model file')

    m = config['MotionModel']
    if not m: return None

    # set some standard params
    model.name = m['name'].encode('utf-8')
    model.dt = m['dt']
    model.measurements = m['measurements']
    model.states = m['states']
    model.accuracy = m['accuracy']
    model.prob_not_assign = m['prob_not_assign']
    model.max_lost = m['max_lost']

    for matrix in matrices:
        if 'sigma' in m[matrix]:
            sigma = m[matrix]['sigma']
        else:
            sigma = 1.0
        m_data = np.matrix(m[matrix]['matrix'],dtype='float')
        setattr(model, matrix, m_data*sigma)

    # call the reshape function to set the matrices to the correct shapes
    model.reshape()

    return model



# def read_object_model(filename):
def read_object_model(config):
    """ read_object_model

    Read in a object model description file and return a dictionary containing
    the appropriate parameters.

    Object models can be described using JSON format, with a basic structure
    as follows:

        {
          "ObjectModel":{
            "name": "UniformState",
            "states": 1,
            "transition": {
              "matrix": [1] }
              ...
            }
        }

    Matrices are flattened JSON arrays.

    Args:
        filename: a JSON file describing the object model.

    Returns:
        model: a core.ObjectModel instance for passing to BayesianTracker

    Notes:
        Note that the matrices are stored as 1D matrices here. In the future,
        this could form part of a Python only object model.

        TODO(arl): More parsing of the data/reshaping arrays. Raise an
        appropriate error if there is something wrong with the model definition.
    """

    m = config['ObjectModel']
    if not m: return None

    matrices = frozenset(['transition','emission','start'])
    model = core.ObjectModel()

    if 'ObjectModel' not in list(config.keys()):
        raise ValueError('Not a valid object model file')

    # set some standard params
    model.name = m['name'].encode('utf-8')
    model.states = m['states']

    for matrix in matrices:
        m_data = np.matrix(m[matrix]['matrix'],dtype='float')
        setattr(model, matrix, m_data)

    # call the reshape function to set the matrices to the correct shapes
    model.reshape()

    return model


def crop_volume(objects, volume=constants.VOLUME):
    """ Return a list of objects that fall within a certain volume """
    axes = ['x','y','z','t']
    within = lambda o: all([getattr(o, a)>=v[0] and getattr(o, a)<=v[1] for a,v in zip(axes, volume)])
    return [o for o in objects if within(o)]


def import_HDF(*args, **kwargs):
    """ Import the HDF data. """
    from . import dataio
    return dataio.import_HDF(*args, **kwargs)

def import_JSON(filename):
    """ Import JSON data """
    raise DeprecationWarning('Use dataio.import_JSON instead')


def build_trees(tracks, update_tracks=True):
    """ Build lineage trees and update the track relationships """
    raise DeprecationWarning('btrack.optimize.lineage deprecated')
    # from btrack.optimise import lineage
    # tree = lineage.LineageTree(tracks)
    # return tree.create(update_tracks=update_tracks)




if __name__ == "__main__":
    pass



#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:     BayesianTracker
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
#-------------------------------------------------------------------------------


__author__ = "Alan R. Lowe"
__email__ = "a.lowe@ucl.ac.uk"

import re
import os

import numpy as np
import matplotlib.pyplot as plt

from matplotlib.collections import LineCollection
import matplotlib.patheffects as PathEffects
from mpl_toolkits.mplot3d.art3d import Line3DCollection
from itertools import product, combinations
from matplotlib.axes import Axes

from . import constants

def _draw_cube(ax, box):
    """ draw_cube

    Plots a cube of dimensions box into the current axes.

    Args:
        ax: a matplotlib3d axes object
        box: dimensions of the box e.g. [(0,1),(0,1),(0,1)]
    Notes:
        Note that the z-order of plotting is not correct.
        DONE(arl): make non cube (i.e. cuboid) shapes.
    """
    if not isinstance(ax, Axes):
        raise TypeError('ax argument must be a matpotlib axis')

    edges = combinations(np.array(list(product(box[0],box[1],box[2]))), 2)

    for s, e in edges:
        if any([np.sum(np.abs(s-e))==box[d][1]-box[d][0] for d in range(3)]):
            ax.plot(*list(zip(s,e)), linestyle=':', color='b')



def _check_plot_order(order):
    """ check_plot_order

    Check that the plotting order is properly formatted. The plotting order is
    specified as a string (e.g. 'xyt') which denotes the axes to plot. Plots
    can either be of two (e.g. 'xz') or three dimensions (e.g. 'xyz').

    Args:
        order: a string containing the plotting order

    Returns:
        DIMS: the number of plotting dimensions if the plot order is properly
            specified, otherwise raises a ValueError
    """
    if not isinstance(order, str):
        raise ValueError('Order must be specified as a string, e.g. xyz')

    # check that the string contains plottable axes
    order_ok = not bool(re.compile(r'[^xyzt]').search(order))
    if not order_ok:
        raise ValueError('Order string is incorrectly specified: {0:s}'
                        '. Should contain one of xyzt.'.format(order))

    # check that we have the correct number of dimensions
    DIMS = len(order)
    if DIMS<2 or DIMS>3:
        raise ValueError('Plot dimensions must be 2D or 3D.')

    return DIMS



def plot_tracks(tracks,
                agents=[],
                lw=1.,
                color_by_type = False,
                terminii=False,
                tail=None,
                box=None,
                order='xyz',
                kalman=False,
                cmap=plt.get_cmap('viridis'),
                labels=False,
                title='BayesianTracker output'):
    """ plot_tracks

    Plot tracks using matplotlib/matplotlib3d. Uses linecollections to speed up
    plotting of many tracks. Two and three dimensional plots can be achieved
    by specifying the order parameter, e.g. order='xy' or 'xyt'...

    Args:
        tracks: a list of tracklets
        order: plot order, default is xyz, but could be xyt etc..
        tail: limit the plotting to the last n timepoints
        box: plot a box to show the limits of the volume
        terminii: bool to plot markers at the start and end of tracks
        kalman: plot the output of the kalman filter
        labels: bool to plot track IDs as labels
        color_by_type: bool to plot track types in different colours

    Returns:
        None

    Notes:
        DONE(arl): parse the order properly to elimate errors
        TODO(arl): implement limits on plotting
        TODO(arl): box plotting in 2D/3D

    """
    DIMS = _check_plot_order(order)

    fig = plt.figure(figsize=(16,16))

    # set up fiddly plot functions
    if DIMS == 3:
        ax = fig.add_subplot(111, projection='3d')
    else:
        ax = fig.add_subplot(111)

    # if the color by type flag is set, partition the tracks according to type:
    if color_by_type:
        track_types = [t.type for t in tracks]


    # use a color map
    colors_rgb = [cmap(int(i)) for i in np.linspace(0,255,16)]
    p_args = {'colors_rgb':colors_rgb, 'order':order, 'labels':labels, 'lw':lw}

    # render the agents
    # _render_agents(ax, agents, **p_args)

    # render the tracks
    _render_tracks(ax, tracks, **p_args)

    # plot a box of the appropriate dimensions
    if box is not None and DIMS == 3:
        box_dims = 'xyzt'
        this_box = [box[box_dims.index(dim)] for dim in order]
        _draw_cube(ax, this_box)

    ax.set_xlabel(order[0])
    ax.set_ylabel(order[1])
    if DIMS == 3:
        ax.set_zlabel(order[2])
        # ax.set_zlim(0, 100)
    else:
        ax.autoscale('tight')

    # plt.axis('image')
    plt.title(title)
    plt.show()



def _render_tracks(ax, tracks, colors_rgb=[], order='xyz', labels=False,
                   terminii=False, lw=1., use_alpha=False):
    """ render_tracks

    Render the tracks on a given set of axes using matplotlib/matplotlib3d.
    Uses linecollections to speed up plotting of many tracks. Two and three
    dimensional plots can be achieved by specifying the order parameter,
    e.g. order='xy' or 'xyt'...
    """
    DIMS = _check_plot_order(order)

    lines, clrs = [], []

    # set up fiddly plot functions
    if DIMS == 3:
        lineplot = lambda l, c: Line3DCollection(l, colors=c)
        addline = lambda ax, l: ax.add_collection3d(l)
    else:
        lineplot = lambda l, c: LineCollection(l, colors=c)
        addline = lambda ax, l: ax.add_collection(l)

    for track in tracks:

        if len(track) < 2: continue

        p_order = [getattr(track, order[i]) for i in range(DIMS)]
        segments = list(zip(*p_order))
        t_color = colors_rgb[track.ID % (len(colors_rgb)-1)]

        lines.append(segments)
        clrs.append(t_color)

        # plot text labels on the tracks
        if labels:
            # set up the plotting arguments
            l_args = [getattr(track, order[i])[-1] for i in range(DIMS)]
            l_args = l_args + [str(track.ID), None]
            # plot the text label with an outline
            ax.text(*l_args, color='k',
                path_effects=[PathEffects.withStroke(linewidth=0.5,
                foreground=t_color)])

        # TODO(arl): add the terminus plotting
        if terminii: pass

    lc = lineplot(lines, clrs)
    lc.set_linewidth(lw)

    addline(ax, lc)






if __name__ == "__main__":
    pass



import numpy as np
from . import btypes

ID_COUNTER = 1



def new_vector(radius=1., theta=0., phi=0.):
        x = lambda r, t, p: r*np.sin(t)*np.cos(p)
        y = lambda r, t, p: r*np.sin(t)*np.sin(p)
        z = lambda r, t, p: r*np.cos(t)
        vect = np.array([x(radius, theta, phi),
                         y(radius, theta, phi),
                         z(radius, theta, phi)])
        return vect


def new_t_obj(position, frame, label):
    """ get a bayesiantracker object from an observation """
    global ID_COUNTER
    obj = btypes.PyTrackObject()
    obj.ID = ID_COUNTER
    obj.x = position[0]
    obj.y = position[1]
    obj.z = position[2]
    obj.t = frame
    obj.dummy = False
    obj.label = label
    obj.probability = np.zeros((5,),dtype='float')
    ID_COUNTER+=1
    return obj


class AgentManager(object):
    def __init__(self,
                 num_agents=100,
                 periodic_boundary=False,
                 noise=0.01,
                 box=1024):

        self.num_agents = num_agents
        self.periodic_boundary = periodic_boundary
        self.noise = noise
        self.box = box

        self.__current_ID = 0

        self.agents = []


    def simulate(self, num_frames):
        """ run the simulation """
        agents = [Agent(ID=self.get_new_ID(),
                  noise=self.noise,
                  periodic_boundary=self.periodic_boundary,
                  position=(512.,512.,512.),
                  frame=0,
                  box=self.box) for a in range(self.num_agents)]

        to_add = []

        # iterate over the frames
        for i in range(num_frames):

            for agent in agents:
                ret = agent.update()
                if ret == 2:
                    to_add = self.mitosis(agent)

            # append any new agents here
            while to_add:
                agents.append(to_add.pop(0))

        self.agents = agents


    def mitosis(self, agent):
        """ perform a mitosis """
        current_pos = agent.position
        r = np.random.randint(5,20)
        theta = np.random.randn()*np.pi
        phi = np.random.randn()*np.pi
        v = new_vector(radius=r, theta=theta, phi=phi)

        frame = agent.frames[-1]+1

        new_pos = [current_pos-v, current_pos+v]
        new_agents = [Agent(ID=self.get_new_ID(),
                            noise=self.noise,
                            periodic_boundary=self.periodic_boundary,
                            frame=frame+1,
                            position=p.tolist(),
                            box=self.box) for p in new_pos]

        print("Mitosis {}->{},{}".format(agent.ID, new_agents[0].ID, new_agents[1].ID,))

        return new_agents


    def get_new_ID(self):
        """ get a new ID """
        self.__current_ID+=1
        return self.__current_ID


    def get_data(self):
        """ get all of the data """
        data = []
        for agent in self.agents:
            data += agent.get_objects()
        return data

class Agent(object):
    """ Agent

    An Agent class that can be used to simulate a moving object in a field of
    view.

    Args:
        noise: normally distributed noise in speed and angle(s)
        periodic_boundary: a bool which causes agents to wrap around edges of
            bounding box
        box: size of initialisation/bounding box
        speed: agent speed

    """
    def __init__(self,
                 ID=0,
                 noise=.1,
                 periodic_boundary=False,
                 position=[],
                 frame=0,
                 box=1024):

        self.ID = ID

        self.box = box
        self.periodic_boundary = periodic_boundary

        if not position:
            self.position = np.random.randint(0,high=self.box, size=(3,))
        else:
            self.position = np.array(position)

        self.theta = np.random.randn(2,)*np.pi # az and elev
        self.speed = 1.
        self.noise = noise

        self.frame = frame

        # position, label=3 (anaphase for starting track)
        self.trajectory = [self.position]
        self.labels = [3]
        self.frames = [frame]

        self.predicted = None

        # sampling noise
        self.sample_noise = 0.1
        self.sample_iter = 0

        # active track?
        self.active = True

        # division?
        self.division_time = 300


    @property
    def division_clock(self):
        return self.frames[-1]-self.frames[0]

    def update(self):
        if not self.active:
            return None

        t = self.theta + np.random.randn(2,)*self.noise
        s = max(self.speed + np.random.randn()*self.noise, 0.)
        # v = [np.cos(t[0])*np.cos(t[1]), np.sin(t[0])*np.cos(t[1]), np.sin(t[1])]
        v = new_vector(radius=s, theta=t[0], phi=t[1])
        v[-1] = 0.
        self.position = self.position + v

        if self.periodic_boundary:
            self.position = np.mod(self.position, self.box)

        if self.divide():
            label = 2 # metaphase
        else:
            label = 0 # interphase

        self.frames.append(self.frames[-1]+1)
        self.trajectory.append(self.position)
        self.labels.append(label)
        self.theta = t

        return label



    def divide(self):
        if self.division_clock >= self.division_time+np.random.randn()*1:
            self.active = False
            return True
        return False


    def get_objects(self):
        objects = []
        for i in range(len(self)):
            o = new_t_obj(self.trajectory[i],
                          self.frames[i],
                          self.labels[i])
            objects.append(o)
        return objects


    def __len__(self): return len(self.trajectory)


    @property
    def x(self): return self.coords(0)
    @property
    def y(self): return self.coords(1)
    @property
    def z(self): return self.coords(2)

    def coords(self, axis=0):
        return [self.trajectory[i][axis] for i in range(len(self.trajectory))]


if __name__ == "__main__":
    pass



"""BayesianTracker (`btrack`) is a multi object tracking algorithm, specifically
used to reconstruct trajectories in crowded fields.  New observations are
assigned to tracks by evaluating the posterior probability of each potential
linkage from a Bayesian belief matrix for all possible linkages. """

from setuptools import setup
from setuptools import find_packages
# from setuptools.command.install import install

def get_install_required():
    with open("./requirements.txt", "r") as reqs:
        requirements = reqs.readlines()
    return [r.rstrip() for r in requirements]

def get_version():
    with open("./btrack/VERSION.txt" ,"r") as ver:
        version = ver.readline()
    return version.rstrip()

DESCRIPTION = 'A framework for Bayesian multi-object tracking'
LONG_DESCRIPTION = __doc__


setup(name='btrack',
      version=get_version(),
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      author='Alan R. Lowe',
      author_email='a.lowe@ucl.ac.uk',
      url='https://github.com/quantumjot/BayesianTracker',
      packages=find_packages(),
      package_data={'btrack': ['libs/libtracker*', 'VERSION.txt']},
      install_requires=get_install_required(),
      python_requires='>=3.6',
      license='LICENSE.md',
      classifiers=['Programming Language :: C++',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 3 :: Only',
                   'Topic :: Scientific/Engineering',
                   'Topic :: Scientific/Engineering :: Bio-Informatics',
                   'Topic :: Scientific/Engineering :: Image Recognition'])



#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:     BayesianTracker
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
#-------------------------------------------------------------------------------


__author__ = "Alan R. Lowe"
__email__ = "code@arlowe.co.uk"

import re
import os
import csv
import h5py
import json
import logging

# import core
from . import btypes
from . import constants

import numpy as np

from collections import OrderedDict
from functools import wraps



# get the logger instance
logger = logging.getLogger('worker_process')


class _PyTrackObjectFactory:
    def __init__(self):
        self.reset()

    def get(self, txyz, label=None, obj_type=0):
        """ get an instatiated object """
        assert(isinstance(txyz, np.ndarray))
        assert(txyz[0] >= 0.) # assert that time is always positive!
        if label is not None:
            if isinstance(label, int):
                class_label = label
                probability = np.zeros((1,))
            else:
                class_label = label[0].astype(np.uint32)
                probability = label[1:].astype(np.float32)
        else:
            class_label = constants.States.NULL.value
            probability = np.zeros((1,))

        new_object = btypes.PyTrackObject()
        new_object.ID = self._ID
        new_object.t = txyz[0].astype(np.uint32)
        new_object.x = txyz[1].astype(np.float32)
        new_object.y = txyz[2].astype(np.float32)
        new_object.z = txyz[3].astype(np.float32)
        new_object.dummy = False
        new_object.label = class_label          # from the classifier
        new_object.probability = probability
        new_object.type = int(obj_type)

        self._ID += 1
        return new_object

    def reset(self):
        self._ID = 0

# instatiate the factory
ObjectFactory = _PyTrackObjectFactory()


def import_JSON(filename):
    """ generic JSON importer for localisations from other software """
    with open(filename, 'r') as json_file:
        data = json.load(json_file)
    objects = []
    for ID, _obj in data.items():
        txyz = np.array([_obj[k] for k in ['t','x','y','z']])
        obj = ObjectFactory.get(txyz, label=int(_obj['label']))
        objects.append(obj)
    return objects


def import_CSV(filename):
    """ import from a CSV file

    CSV file should have one of the following formats:

    t, x, y
    t, x, y, label
    t, x, y, z
    t, x, y, z, label

    """
    objects = []
    with open(filename, 'r') as csv_file:
        csvreader = csv.DictReader(csvfile, delimiter=' ', quotechar='|')
        for row in csvreader:
            txyz = np.zeros((1,4), dtype=np.float32)
            txyz[:,(0, 1, 2)] = [row[k] for k in ('t', 'x', 'y')]
            if 'z' in row:
                txyz[:,3] = row['z'] # if we have z info
            if 'label' in row:
                label = int(row['label'])
            else:
                label = None

            objects.append(ObjectFactory.get(txyz, label=label))
    return objects




def export_delegator(filename, tracker, obj_type=None, filter_by=None):
    """ Export data from the tracker using the appropriate exporter """
    # assert(isinstance(tracker, BayesianTracker))
    assert(isinstance(filename, str))

    export_dir, export_fn = os.path.split(filename)
    _, ext = os.path.splitext(filename)

    assert(os.path.exists(export_dir))

    if ext == '.json':
        raise DeprecationWarning('JSON export is deprecated')
    elif ext == '.mat':
        raise DeprecationWarning('MATLAB export is deprecated')
    elif ext == '.csv':
        export_CSV(filename, tracker.tracks, obj_type=obj_type)
    elif ext in ('.hdf', '.hdf5', '.h5'):
        _export_HDF(filename, tracker, obj_type=obj_type, filter_by=filter_by)
    else:
        logger.error(f'Export file format {ext} not recognized.')





def check_track_type(tracks):
    # return isinstance(tracks[0], btypes.Tracklet)
    return all([isinstance(t, btypes.Tracklet) for t in tracks])



def export_CSV(filename: str,
               tracks: list,
               properties: list = constants.DEFAULT_EXPORT_PROPERTIES,
               obj_type=None):
    """ export the track data as a simple CSV file """

    if not tracks:
        logger.error(f'No tracks found when exporting to: {filename}')
        return

    if not check_track_type(tracks):
        logger.error('Tracks of incorrect type')

    logger.info(f'Writing out CSV files to: {filename}')
    export_track = np.vstack([t.to_array(properties) for t in tracks])

    with open(filename, 'w', newline='') as csvfile:
        csvwriter = csv.writer(csvfile, delimiter=' ')
        csvwriter.writerow(properties)
        for i in range(export_track.shape[0]):
            csvwriter.writerow(export_track[i,:].tolist())




def export_LBEP(filename: str, tracks: list):
    """ export the LBEP table described here:
    https://public.celltrackingchallenge.net/documents/
        Naming%20and%20file%20content%20conventions.pdf
    """
    if not tracks:
        logger.error(f'No tracks found when exporting to: {filename}')
        return

    if not check_track_type(tracks):
        logger.error('Tracks of incorrect type')

    tracks.sort(key=lambda t: t.ID)
    if not filename.endswith('.txt'): filename+='.txt'
    with open(filename, 'w') as lbep_file:
        logger.info(f'Writing LBEP file: {filename}...')
        for track in tracks:
            lbep = f'{track.ID} {track.t[0]} {track.t[-1]} {track.parent}'
            lbep_file.write(f'{lbep}\n')


def _export_HDF(filename: str,
               tracker,
               obj_type=None,
               filter_by: str = None):
    """ export to HDF """
    with HDF5FileHandler(filename, read_write='a') as hdf:
        hdf.write_tracks(tracker, obj_type=obj_type, f_expr=filter_by)



def h5check_property_exists(property):
    """ Wrapper for hdf handler to make sure a property exists """
    def func(fn):
        @wraps(fn)
        def wrapped_handler_property(*args, **kwargs):
            self = args[0]
            assert(isinstance(self, HDF5FileHandler))
            if property not in self._hdf:
                logger.error(f'{property.capitalize()} not found in {self.filename}')
                return None
            return fn(*args, **kwargs)
        return wrapped_handler_property
    return func



class HDF5FileHandler:
    """ HDF5FileHandler

    Generic HDF5 file hander for reading and writing datasets. This is
    inter-operable between segmentation, tracking and analysis code.

    LBEPR is a modification of the LBEP format to also include the root node
    of the tree.

        I - number of objects
        J - number of frames
        K - number of tracks

    Added generic filtering to object retrieval, e.g.
        obj = handler.filtered_objects('flag==1')
        retrieves all objects if there is an object['flag'] == 1

    Basic format of the HDF file is:
        segmentation/
            images          - (J x h x w) uint8 images of the segmentation
        objects/
            obj_type_1/
                coords      - (I x 5) [t, x, y, z, object_type]
                labels      - (I x D) [label, (softmax scores ...)]
                map         - (J x 2) [start_index, end_index] -> coords array
            ...
        tracks/
            obj_type_1/
                tracks      - (I x 1) [index into coords]
                dummies     - similar to coords, but for dummy objects
                map         - (K x 2) [start_index, end_index] -> tracks array
                LBEPRG      - (K x 6) [L, B, E, P, R, G]
                fates       - (K x n) [fate_from_tracker, ...future_expansion]
            ...

    Notes:
        NOTE(arl): the final slice [:] reads the whole file in one go,
        since we are unlikely to have more objects than memory and we
        need to load them all anyway.

        NOTE(arl): should dummies be moved to coords?
    """

    def __init__(self, filename, read_write='r'):
        self.filename = filename
        logger.info(f'Opening HDF file: {self.filename}')
        self._hdf = h5py.File(filename, read_write)
        self._states = list(constants.States)
        self._f_expr = None # DO NOT USE

    @property
    def object_types(self):
        return list(self._hdf['objects'].keys())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        if not self._hdf: return
        logger.info(f'Closing HDF file: {self.filename}')
        self._hdf.close()

    @property
    def objects(self):
        """ Return the objects in the file """
        return self.filtered_objects()

    @h5check_property_exists('objects')
    def filtered_objects(self, f_expr=None, obj_types=None):
        """ return a filtered list of objects based on metadata.
        f_expr should be of the format 'flag==1'
        """

        objects = []
        if obj_types is None:
            obj_types = self.object_types
        else:
            assert(isinstance(obj_types, list))
            assert(all([o in self.object_types for o in obj_types]))

        for ci, c in enumerate(obj_types):
            # read the whole dataset into memory
            txyz = self._hdf['objects'][c]['coords'][:]
            if 'labels' not in self._hdf['objects'][c]:
                logger.warning('Labels missing from objects in HDF file')
                labels = np.zeros((txyz.shape[0], 6))
            else:
                labels = self._hdf['objects'][c]['labels'][:]
            idx = range(txyz.shape[0])      # default filtering uses all objects

            # note that this doesn't do much error checking at the moment
            if f_expr is not None:
                assert(isinstance(f_expr, str))
                pattern = '(?P<name>\w+)(?P<op>[\>\<\=]+)(?P<cmp>[0-9]+)'
                m = re.match(pattern, f_expr)
                f_eval = f'x{m["op"]}{m["cmp"]}' # e.g. x > 10

                if m['name'] in self._hdf['objects'][c]:
                    data = self._hdf['objects'][c][m['name']][:]
                    idx = [i for i, x in enumerate(data) if eval(f_eval)]
                else:
                    logger.warning(f'Cannot filter objects by {m["name"]}')

            assert(txyz.shape[0] == labels.shape[0])
            logger.info(f'Loading {c} {txyz.shape} ({len(idx)} filtered: {f_expr})...')
            obj = [ObjectFactory.get(txyz[i,:], label=labels[i,:], obj_type=ci+1) for i in idx]
            objects += obj
        return objects

    def write_tracks(self, tracker, obj_type=None, f_expr=None):
        """ Write tracks to HDF file """
        if not tracker.tracks:
            logger.error(f'No tracks found when exporting to: {self.filename}')
            return

        assert(obj_type in self.object_types)

        tracks = tracker.refs
        dummies = tracker.dummies
        lbep_table = np.stack(tracker.lbep, axis=0).astype(np.int32)

        # sanity check
        assert(lbep_table.shape[0] == len(tracks))

        logger.info(f'Writing tracks to HDF file: {self.filename}')
        hdf_tracks = np.concatenate(tracks, axis=0)

        hdf_frame_map = np.zeros((len(tracks),2), dtype=np.int32)
        for i, track in enumerate(tracks):
            if i > 0:
                offset = hdf_frame_map[i-1,1]
            else: offset = 0
            hdf_frame_map[i,:] = np.array([0, len(track)]) + offset

        if 'tracks' not in self._hdf:
            self._hdf.create_group('tracks')

        if obj_type in self._hdf['tracks']:
            logger.warning(f'Removing {obj_type} from HDF file. ')
            del self._hdf['tracks'][obj_type]

        grp = self._hdf['tracks'].create_group(obj_type)
        grp.create_dataset('tracks', data=hdf_tracks, dtype='int32')
        grp.create_dataset('map', data=hdf_frame_map, dtype='uint32')

        # if we have used the f_expr we can save it as an attribute here
        if f_expr is not None and isinstance(f_expr, str):
            grp.attrs['f_expr'] = f_expr

        # also save the version number as an attribute
        grp.attrs['version'] = constants.get_version()

        # write out dummies
        if dummies:
            logger.info(f'Writing dummies to HDF file: {self.filename}')
            o = self.object_types.index(obj_type) + 1
            txyz = np.stack([[d.t, d.x, d.y, d.z, o] for d in dummies], axis=0)
            grp.create_dataset('dummies', data=txyz, dtype='float32')

        # write out the LBEP table
        logger.info(f'Writing LBEPR to HDF file: {self.filename}')
        grp.create_dataset('LBEPR', data=lbep_table, dtype='int32')

        # write out cell fates
        logger.info(f'Writing track fates to HDF file: {self.filename}')
        fate_table = np.stack([t.fate.value for t in tracker.tracks], axis=0)
        grp.create_dataset('fates', data=fate_table, dtype='int32')

    @property
    @h5check_property_exists('tracks')
    def tracks(self):
        """ Return the tracks in the file """
        dummies, ret = [], []

        # make an object factory
        factory = _PyTrackObjectFactory()

        for ci, c in enumerate(self._hdf['tracks'].keys()):
            logger.info(f'Loading tracks: {c}...')
            track_map = self._hdf['tracks'][c]['map'][:]
            track_refs = self._hdf['tracks'][c]['tracks'][:]
            lbep = self._hdf['tracks'][c]['LBEPR'][:]
            fates = self._hdf['tracks'][c]['fates'][:]

            # if there are dummies, make new dummy objects
            if 'dummies' in self._hdf['tracks'][c]:
                dummies = self._hdf['tracks'][c]['dummies'][:]
                n_dummies = dummies.shape[0]
                dobj = [factory.get(dummies[i,:]) for i in range(n_dummies)]
                for d in dobj: d.dummy = True

            # TODO(arl): this needs to be stored in the HDF folder
            if 'f_expr' in self._hdf['tracks'][c].attrs:
                f_expr = self._hdf['tracks'][c].attrs['f_expr']
            elif self._f_expr is not None:
                f_expr = self._f_expr
            else:
                f_expr = None

            obj = self.filtered_objects(f_expr=f_expr, obj_types=[c])

            def get_txyz(ref):
                if ref>=0: return obj[ref]
                return dobj[abs(ref)-1] # references are -ve for dummies

            tracks = []
            for i in range(track_map.shape[0]):
                idx = slice(*track_map[i,:].tolist())
                refs = track_refs[idx]
                track = btypes.Tracklet(lbep[i,0], list(map(get_txyz, refs)))
                track.parent = lbep[i,3]    # set the parent and root of tree
                track.root = lbep[i,4]
                if lbep.shape[1] > 5: track.generation = lbep[i,5]
                track.fate = constants.Fates(fates[i]) # restore the track fate
                tracks.append(track)

            # once we have all of the tracks, populate the children
            to_update = {}
            for track in tracks:
                if not track.is_root:
                    parents = filter(lambda t: t.ID == track.parent, tracks)
                    for parent in parents:
                        if parent not in to_update:
                            to_update[parent] = []
                        to_update[parent].append(track.ID)

            # sanity check, can be removed at a later date
            assert all([len(children)<=2 for children in to_update.values()])

            # add the children to the parent
            for track, children in to_update.items():
                track.children = children

            ret.append(tracks)

        return ret

    @property
    def segmentation(self):
        raise NotImplementedError

    @property
    @h5check_property_exists('tracks')
    def lbep(self):
        logger.info('Loading LBEPR tables...')
        return [self._hdf['tracks'][k]['LBEPR'][:] for k in self._hdf['tracks']]






if __name__ == "__main__":
    pass
